// lab1.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "memory.h"
#include "time.h"

#include "omp.h"

#include <IL/ilut.h>
#include <IL/ilu.h>

#include "emmintrin.h"
#include "nmmintrin.h"

#define FILTER_W 5
#define FILTER_H 5

void conv_filter(int imgHeight, int imgWidth, int imgHeightF, int imgWidthF,
				 int imgFOfssetH, int imgFOfssetW,
				 float *filter, float *imgFloatSrc, float *imgFloatDst)
{
	for (int row=imgFOfssetH; row<(imgHeight+imgFOfssetH); row++)
	{
		for (int col=imgFOfssetW; col<(imgWidth+imgFOfssetW); col++)
		{
			int pixel_dst = (row*imgWidthF+col)*4;
			for (int rgba=0; rgba<4; rgba++)
			{
				float fval = 0;
				for (int fy=0; fy<FILTER_H; fy++)
				{
					for (int fx=0; fx<FILTER_W; fx++)
					{
						int pixel_src = ((row+fy-imgFOfssetH)*imgWidthF + col+fx-imgFOfssetW)*4;
						fval = fval + *(filter+fy*FILTER_W+fx) * *(imgFloatSrc+pixel_src+rgba);
					}
				}
				if (fval>255.0) fval=255.0;
				if (fval<0.0)   fval=0.0;
				*(imgFloatDst+pixel_dst+rgba) = fval;
			}
		}
	}
}

void conv_grey(int imgHeightF, int imgWidthF,
			   float *imgFloatSrc, float *imgFloatDst)
{
	for (int row=0; row<imgHeightF; row++)
	{
		for (int col=0; col<imgWidthF; col++)
		{
			int pixel = (row*imgWidthF+col)*4;
			float avg;
			avg = *(imgFloatSrc+pixel+0) + *(imgFloatSrc+pixel+1) + *(imgFloatSrc+pixel+2);
			*(imgFloatDst+pixel+0) = avg;
			*(imgFloatDst+pixel+1) = avg;
			*(imgFloatDst+pixel+2) = avg;
		}
	}
}

void _tmain(int argc, _TCHAR* argv[])
{
	ilInit(); iluInit();
	ILboolean ret;
	ILuint ilImg=0;
	ilGenImages(1, &ilImg);
	ilBindImage(ilImg);
	ret = ilLoadImage((wchar_t*)("test.bmp"));
	ILubyte* imgData = ilGetData(); 

	int imgWidth = ilGetInteger(IL_IMAGE_WIDTH);
	int imgHeight = ilGetInteger(IL_IMAGE_HEIGHT);
	ILint imgOrigin = ilGetInteger(IL_ORIGIN_MODE);

	float *imgFloat;
	int imgWidthF = imgWidth+FILTER_W-1;
	int imgHeightF = imgHeight+FILTER_H-1;
	int imgFOfssetW = (FILTER_W-1)/2;
	int imgFOfssetH = (FILTER_H-1)/2;
	imgFloat   = (float *)(_aligned_malloc(4*imgWidthF*imgHeightF*sizeof(float), 16));
    int row, col;

	for (row=0; row<imgHeightF; row++)
	{
		for (col=0; col<imgWidthF;col++)
		{
			int pixel = (row*imgWidthF + col)*4;
			*(imgFloat+pixel+0) = 0.0;
			*(imgFloat+pixel+1) = 0.0;
			*(imgFloat+pixel+2) = 0.0;
			*(imgFloat+pixel+3) = 0.0;
		}
	}

	for (row=0; row<imgHeight; row++)
	{
		for (col=0; col<imgWidth;col++)
		{
			int pixel_dst = ((row+imgFOfssetH)*imgWidthF + (col+imgFOfssetW))*4;
			int pixel_src = (row*imgWidth + col)*3;
			*(imgFloat+pixel_dst+0) = (float)(*(imgData+pixel_src+0));
			*(imgFloat+pixel_dst+1) = (float)(*(imgData+pixel_src+1));
			*(imgFloat+pixel_dst+2) = (float)(*(imgData+pixel_src+2));
			*(imgFloat+pixel_dst+3) = 0.0;
		}
	}


// IMAGE PROCESSING
	clock_t t_s0, t_f0;
	double duration;
	
	float *imgFloatRes;
    imgFloatRes = (float *)(_aligned_malloc(4*imgWidthF*imgHeightF*sizeof(float), 16));

	float filter_gauss[] = {2.0,  4.0,  5.0,  4.0,  2.0,
	                        4.0,  9.0, 12.0,  9.0,  4.0,
	                        5.0, 12.0, 15.0, 12.0,  5.0,
	                        4.0,  9.0, 12.0,  9.0,  4.0,
                            2.0,  4.0,  5.0,  4.0,  2.0};
    float fsum = 0;
	for (int f=0; f<FILTER_W*FILTER_H; f++)
		fsum = fsum + filter_gauss[f];
	for (int f=0; f<FILTER_W*FILTER_H; f++)
		*(filter_gauss+f) = filter_gauss[f]/(fsum);


	t_s0 = clock();
for (int r=0;r<100;r++)
{
/*
	conv_filter(imgHeight, imgWidth, imgHeightF, imgWidthF,
				imgFOfssetH, imgFOfssetW,
				filter_gauss, imgFloat, imgFloatRes);
*/
/*
	conv_grey(imgHeightF, imgWidthF,
			  imgFloatRes, imgFloat);
*/

	// Laplace edge detection

	float filter_laplace[] = {-1.0, -1.0, -1.0, -1.0, -1.0,
		                      -1.0, -1.0, -1.0, -1.0, -1.0,
		                      -1.0, -1.0, 24.0, -1.0, -1.0,
		                      -1.0, -1.0, -1.0, -1.0, -1.0,
		                      -1.0, -1.0, -1.0, -1.0, -1.0};

	conv_filter(imgHeight, imgWidth, imgHeightF, imgWidthF,
				imgFOfssetH, imgFOfssetW,
				filter_laplace, imgFloat, imgFloatRes);

}
	t_f0 = clock();
	duration = (double)(t_f0 - t_s0)/CLOCKS_PER_SEC;
	printf("Processing: %4.4f\n", duration);

	float *imgFloatWrite;
	imgFloatWrite = imgFloatRes;

// IMAGE PROCESSING END


	for (row=0; row<imgHeight; row++)
	{
		for (col=0; col<imgWidth;col++)
		{
			int pixel_src = ((row+imgFOfssetH)*imgWidthF + (col+imgFOfssetW))*4;
			int pixel_dst = (row*imgWidth + col)*3;
			*(imgData+pixel_dst+0) = (ILubyte)(*(imgFloatWrite+pixel_src+0));
			*(imgData+pixel_dst+1) = (ILubyte)(*(imgFloatWrite+pixel_src+1));
			*(imgData+pixel_dst+2) = (ILubyte)(*(imgFloatWrite+pixel_src+2));
		}
	}


	_aligned_free(imgFloat);
    _aligned_free(imgFloatRes);

	ret = ilSetData(imgData);
	ilEnable(IL_FILE_OVERWRITE);
	ilSaveImage((wchar_t*)("test2.bmp"));
	ilDeleteImages(1, &ilImg);
}

